"""Contains an event manager class."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine, Generator
import inspect
from types import MappingProxyType
from typing import Any, Generic, TypeAlias, TypeVar, overload

from pyplumio.helpers.task_manager import TaskManager

EventCallback: TypeAlias = Callable[[Any], Coroutine[Any, Any, Any]]
_Callable: TypeAlias = Callable[..., Any]

_EventCallbackT = TypeVar("_EventCallbackT", bound=EventCallback)


@overload
def event_listener(name: _Callable, filter: None = None) -> EventCallback: ...


@overload
def event_listener(
    name: str | None = None, filter: _Callable | None = None
) -> _Callable: ...


def event_listener(name: Any = None, filter: _Callable | None = None) -> Any:
    """Mark a function as an event listener.

    This decorator attaches metadata to the function, identifying it
    as a subscriber for the specified event.
    """

    def decorator(func: _EventCallbackT) -> _EventCallbackT:
        # Attach metadata to the function to mark it as a listener.
        event = (
            name
            if isinstance(name, str)
            else func.__qualname__.split("on_event_", 1)[1]
        )
        setattr(func, "_on_event", event)
        setattr(func, "_on_event_filter", filter)
        return func

    if callable(name):
        return decorator(name)
    else:
        return decorator


T = TypeVar("T")


class EventManager(TaskManager, Generic[T]):
    """Represents an event manager."""

    __slots__ = ("_data", "_events", "_callbacks")

    _data: dict[str, T]
    _events: dict[str, asyncio.Event]
    _callbacks: dict[str, list[EventCallback]]

    def __init__(self) -> None:
        """Initialize a new event manager."""
        super().__init__()
        self._data = {}
        self._events = {}
        self._callbacks = {}
        self._register_event_listeners()

    def __getattr__(self, name: str) -> T:
        """Return attributes from the underlying data dictionary."""
        try:
            return self.data[name]
        except KeyError as e:
            raise AttributeError from e

    def _register_event_listeners(self) -> None:
        """Register the event listeners."""
        for event, callback in self.event_listeners():
            filter_func = getattr(callback, "_on_event_filter", None)
            self.subscribe(event, filter_func(callback) if filter_func else callback)

    def event_listeners(self) -> Generator[tuple[str, EventCallback]]:
        """Get the event listeners."""
        for _, callback in inspect.getmembers(self, predicate=inspect.ismethod):
            if event := getattr(callback, "_on_event", None):
                yield (event, callback)

    async def wait_for(self, name: str, timeout: float | None = None) -> None:
        """Wait for the value to become available.

        :param name: Event name or ID
        :type name: str
        :param timeout: Wait this amount of seconds for a data to
            become available, defaults to `None`
        :type timeout: float, optional
        :raise asyncio.TimeoutError: when waiting past specified timeout
        """
        if name not in self.data:
            await asyncio.wait_for(self.create_event(name).wait(), timeout=timeout)

    async def get(self, name: str, timeout: float | None = None) -> T:
        """Get the value by name.

        :param name: Event name or ID
        :type name: str
        :param timeout: Wait this amount of seconds for a data to
            become available, defaults to `None`
        :type timeout: float, optional
        :return: An event data
        :rtype: T
        :raise asyncio.TimeoutError: when waiting past specified timeout
        """
        await self.wait_for(name, timeout=timeout)
        return self.data[name]

    @overload
    def get_nowait(self, name: str, default: None = ...) -> T | None: ...

    @overload
    def get_nowait(self, name: str, default: T) -> T: ...

    def get_nowait(self, name: str, default: Any = None) -> Any:
        """Get the value by name without waiting.

        If value is not available, default value will be
        returned instead.

        :param name: Event name or ID
        :type name: str
        :param default: default value to return if data is unavailable,
            defaults to `None`
        :type default: T, optional
        :return: An event data
        :rtype: T, optional
        """
        try:
            return self.data[name]
        except KeyError:
            return default

    def subscribe(self, name: str, callback: _EventCallbackT) -> _EventCallbackT:
        """Subscribe a callback to the event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, that will be
            awaited on the with the event data as an argument.
        :type callback: Callback
        :return: A reference to the callback, that can be used
            with `EventManager.unsubscribe()`.
        :rtype: Callback
        """
        callbacks = self._callbacks.setdefault(name, [])
        callbacks.append(callback)
        return callback

    def subscribe_once(self, name: str, callback: EventCallback) -> EventCallback:
        """Subscribe a callback to the event once.

        Callback will be unsubscribed after single event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, that will be
            awaited on the with the event data as an argument.
        :type callback: Callback
        :return: A reference to the callback, that can be used
            with `EventManager.unsubscribe()`.
        :rtype: Callback
        """

        async def _call_once(value: Any) -> Any:
            """Unsubscribe callback from the event and calls it."""
            self.unsubscribe(name, _call_once)
            return await callback(value)

        return self.subscribe(name, _call_once)

    def unsubscribe(self, name: str, callback: EventCallback) -> bool:
        """Usubscribe a callback from the event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, previously
            subscribed to an event using ``subscribe()`` or
            ``subscribe_once()`` methods.
        :type callback: Callback
        :return: `True` if callback is found, `False` otherwise.
        :rtype: bool
        """
        if name in self._callbacks and callback in self._callbacks[name]:
            self._callbacks[name].remove(callback)
            return True

        return False

    async def dispatch(self, name: str, value: T) -> None:
        """Call registered callbacks and dispatch the event."""
        if callbacks := self._callbacks.get(name, None):
            for callback in list(callbacks):
                if (result := await callback(value)) is not None:
                    value = result

        self._data[name] = value
        self.set_event(name)

    def dispatch_nowait(self, name: str, value: T) -> None:
        """Call a registered callbacks and dispatch the event without waiting."""
        self.create_task(self.dispatch(name, value))

    async def load(self, data: dict[str, T]) -> None:
        """Load event data."""
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in data.items())
        )

    def load_nowait(self, data: dict[str, T]) -> None:
        """Load event data without waiting."""
        self.create_task(self.load(data))

    def create_event(self, name: str) -> asyncio.Event:
        """Create an event."""
        if name in self.events:
            return self.events[name]

        event = asyncio.Event()
        self._events[name] = event
        return event

    def set_event(self, name: str) -> None:
        """Set an event."""
        if name in self.events:
            self.events[name].set()

    @property
    def events(self) -> dict[str, asyncio.Event]:
        """Return the events."""
        return self._events

    @property
    def data(self) -> MappingProxyType[str, T]:
        """Return the event data."""
        return MappingProxyType(self._data)


__all__ = ["EventCallback", "EventManager", "event_listener"]



"""Contains a simple async cache for caching results of async functions."""

from collections.abc import Awaitable, Callable
from functools import wraps
from types import MappingProxyType
from typing import Any, ParamSpec, TypeVar, cast

T = TypeVar("T")
P = ParamSpec("P")


class AsyncCache:
    """A simple cache for asynchronous functions."""

    __slots__ = ("_cache",)

    _cache: dict[str, Any]

    def __init__(self) -> None:
        """Initialize the cache."""
        self._cache = {}

    async def get(self, key: str, coro: Callable[..., Awaitable[Any]]) -> Any:
        """Get a value from the cache or compute and store it."""
        if key not in self.cache:
            self._cache[key] = await coro()

        return self._cache[key]

    @property
    def cache(self) -> MappingProxyType[str, Any]:
        """Return the internal cache dictionary."""
        return MappingProxyType(self._cache)


# Create a global cache instance
async_cache = AsyncCache()


def acache(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    """Cache the result of an async function."""

    @wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        func_name = f"{func.__module__}.{func.__qualname__}"
        key = f"{func_name}:{args}:{kwargs}"
        return cast(T, await async_cache.get(key, lambda: func(*args, **kwargs)))

    return wrapper


__all__ = ["acache", "async_cache"]



"""Contains schedule decoder."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import dataclass
import datetime as dt
from functools import lru_cache
from typing import Annotated, Any, Final, TypeAlias, get_args

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    STATE_OFF,
    STATE_ON,
    FrameType,
    State,
)
from pyplumio.devices import Device, PhysicalDevice
from pyplumio.frames import Request
from pyplumio.parameters import (
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict, split_byte

ATTR_SCHEDULES: Final = "schedules"
ATTR_SCHEDULE_PARAMETERS: Final = "schedule_parameters"
ATTR_SCHEDULE_SWITCH: Final = "schedule_switch"
ATTR_SCHEDULE_PARAMETER: Final = "schedule_parameter"

SCHEDULE_SIZE: Final = 42  # 6 bytes per day, 7 days total

SCHEDULES: tuple[str, ...] = (
    "heating",
    "water_heater",
    "circulation_pump",
    "boiler_work",
    "boiler_clean",
    "hear_exchanger_clean",
    "mixer_1",
    "mixer_2",
    "mixer_3",
    "mixer_4",
    "mixer_5",
    "mixer_6",
    "mixer_7",
    "mixer_8",
    "mixer_9",
    "mixer_10",
    "thermostat_1",
    "thermostat_2",
    "thermostat_3",
    "circuit_1",
    "circuit_2",
    "circuit_3",
    "circuit_4",
    "circuit_5",
    "circuit_6",
    "circuit_7",
    "panel_1",
    "panel_2",
    "panel_3",
    "panel_4",
    "panel_5",
    "panel_6",
    "panel_7",
    "main_heater_solar",
    "heating_circulation",
    "internal_thermostat",
    "heater",
    "water_heater_2",
    "intake",
    "intake_summer",
)


@dataclass
class ScheduleParameterDescription(ParameterDescription):
    """Represent a schedule parameter description."""


class ScheduleParameter(Parameter):
    """Represents a schedule parameter."""

    __slots__ = ()

    device: PhysicalDevice
    description: ScheduleParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        schedule_name = self.description.name.split("_schedule_", 1)[0]
        return await Request.create(
            FrameType.REQUEST_SET_SCHEDULE,
            recipient=self.device.address,
            data=collect_schedule_data(schedule_name, self.device),
        )


@dataclass(slots=True)
class ScheduleNumberDescription(ScheduleParameterDescription, NumberDescription):
    """Represents a schedule number description."""


class ScheduleNumber(ScheduleParameter, Number):
    """Represents a schedule number."""

    __slots__ = ()

    description: ScheduleNumberDescription


@dataclass(slots=True)
class ScheduleSwitchDescription(ScheduleParameterDescription, SwitchDescription):
    """Represents a schedule switch description."""


class ScheduleSwitch(ScheduleParameter, Switch):
    """Represents a schedule switch."""

    __slots__ = ()

    description: ScheduleSwitchDescription


SCHEDULE_PARAMETERS: list[ScheduleParameterDescription] = [
    description
    for name in SCHEDULES
    for description in [
        ScheduleSwitchDescription(name=f"{name}_{ATTR_SCHEDULE_SWITCH}"),
        ScheduleNumberDescription(name=f"{name}_{ATTR_SCHEDULE_PARAMETER}"),
    ]
]


def collect_schedule_data(name: str, device: Device) -> dict[str, Any]:
    """Return a schedule data collected from the device."""
    return {
        ATTR_TYPE: name,
        ATTR_SWITCH: device.data[f"{name}_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: device.data[f"{name}_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: device.data[ATTR_SCHEDULES][name],
    }


TIME_FORMAT: Final = "%H:%M"

Time: TypeAlias = Annotated[str, "Time string in %H:%M format"]

MIDNIGHT: Final = Time("00:00")
MIDNIGHT_DT = dt.datetime.strptime(MIDNIGHT, TIME_FORMAT)

STEP = dt.timedelta(minutes=30)


def get_time(
    index: int, start: dt.datetime = MIDNIGHT_DT, step: dt.timedelta = STEP
) -> Time:
    """Return time for a specific index."""
    time_dt = start + (step * index)
    return f"{time_dt.hour:02d}:{time_dt.minute:02d}"


@lru_cache(maxsize=10)
def get_time_range(start: Time, end: Time, step: dt.timedelta = STEP) -> list[Time]:
    """Get a time range.

    Start and end boundaries should be specified in %H:%M format.
    Both are inclusive.
    """
    start_dt = dt.datetime.strptime(start, TIME_FORMAT)
    end_dt = dt.datetime.strptime(end, TIME_FORMAT)

    if end_dt == MIDNIGHT_DT:
        # Upper boundary of the interval is midnight.
        end_dt += dt.timedelta(hours=24) - step

    if end_dt <= start_dt:
        raise ValueError(
            f"Invalid time range: start time ({start}) must be earlier "
            f"than end time ({end})."
        )

    seconds = (end_dt - start_dt).total_seconds()
    steps = seconds // step.total_seconds() + 1

    return [get_time(index, start=start_dt, step=step) for index in range(int(steps))]


class ScheduleDay(MutableMapping):
    """Represents a single day of schedule."""

    __slots__ = ("_schedule",)

    _schedule: dict[Time, bool]

    def __init__(self, schedule: dict[Time, bool]) -> None:
        """Initialize a new schedule day."""
        self._schedule = schedule

    def __repr__(self) -> str:
        """Return serializable representation of the class."""
        return f"ScheduleDay({self._schedule})"

    def __len__(self) -> int:
        """Return a schedule length."""
        return self._schedule.__len__()

    def __iter__(self) -> Iterator[Time]:
        """Return an iterator."""
        return self._schedule.__iter__()

    def __getitem__(self, time: Time) -> State:
        """Return a schedule item."""
        state = self._schedule.__getitem__(time)
        return STATE_ON if state else STATE_OFF

    def __delitem__(self, time: Time) -> None:
        """Delete a schedule item."""
        self._schedule.__delitem__(time)

    def __setitem__(self, time: Time, state: State | bool) -> None:
        """Set a schedule item."""
        if state in get_args(State):
            state = True if state == STATE_ON else False
        if isinstance(state, bool):
            self._schedule.__setitem__(time, state)
        else:
            raise TypeError(
                f"Expected boolean value or one of: {', '.join(get_args(State))}."
            )

    def set_state(
        self, state: State | bool, start: Time = MIDNIGHT, end: Time = MIDNIGHT
    ) -> None:
        """Set a schedule interval state."""
        for time in get_time_range(start, end):
            self.__setitem__(time, state)

    def set_on(self, start: Time = MIDNIGHT, end: Time = MIDNIGHT) -> None:
        """Set a schedule interval state to 'on'."""
        self.set_state(STATE_ON, start, end)

    def set_off(self, start: Time = MIDNIGHT, end: Time = MIDNIGHT) -> None:
        """Set a schedule interval state to 'off'."""
        self.set_state(STATE_OFF, start, end)

    @property
    def schedule(self) -> dict[Time, bool]:
        """Return the schedule."""
        return self._schedule

    @classmethod
    def from_iterable(cls: type[ScheduleDay], intervals: Iterable[bool]) -> ScheduleDay:
        """Make schedule day from iterable."""
        return cls({get_time(index): state for index, state in enumerate(intervals)})


@dataclass(slots=True)
class Schedule(Iterable):
    """Represents a weekly schedule."""

    name: str
    device: PhysicalDevice

    sunday: ScheduleDay
    monday: ScheduleDay
    tuesday: ScheduleDay
    wednesday: ScheduleDay
    thursday: ScheduleDay
    friday: ScheduleDay
    saturday: ScheduleDay

    def __iter__(self) -> Iterator[ScheduleDay]:
        """Return list of days."""
        return (
            self.sunday,
            self.monday,
            self.tuesday,
            self.wednesday,
            self.thursday,
            self.friday,
            self.saturday,
        ).__iter__()

    async def commit(self) -> None:
        """Commit a weekly schedule to the device."""
        await self.device.queue.put(
            await Request.create(
                FrameType.REQUEST_SET_SCHEDULE,
                recipient=self.device.address,
                data=collect_schedule_data(self.name, self.device),
            )
        )


class SchedulesStructure(StructureDecoder):
    """Represents a schedule data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_schedule(self, message: bytearray) -> list[list[bool]]:
        """Unpack a schedule."""
        offset = self._offset
        schedule = [
            bit
            for i in range(offset, offset + SCHEDULE_SIZE)
            for bit in split_byte(message[i])
        ]
        self._offset = offset + SCHEDULE_SIZE
        # Split the schedule. Each day consists of 48 half-hour intervals.
        return [schedule[i : i + 48] for i in range(0, len(schedule), 48)]

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        try:
            start = message[offset + 1]
            end = message[offset + 2]
        except IndexError:
            return ensure_dict(data, {ATTR_SCHEDULES: []}), offset

        schedules: list[tuple[int, list[list[bool]]]] = []
        parameters: list[tuple[int, ParameterValues]] = []

        self._offset = offset + 3
        for _ in range(start, start + end):
            index = message[self._offset]
            switch = ParameterValues(
                value=message[self._offset + 1], min_value=0, max_value=1
            )
            parameter = unpack_parameter(message, self._offset + 2)
            self._offset += 5
            schedules.append((index, self._unpack_schedule(message)))
            parameters.append((index * 2, switch))
            if parameter is not None:
                parameters.append((index * 2 + 1, parameter))

        return (
            ensure_dict(
                data, {ATTR_SCHEDULES: schedules, ATTR_SCHEDULE_PARAMETERS: parameters}
            ),
            self._offset,
        )


__all__ = [
    "ATTR_SCHEDULES",
    "ATTR_SCHEDULE_PARAMETERS",
    "ATTR_SCHEDULE_SWITCH",
    "ATTR_SCHEDULE_PARAMETER",
    "Schedule",
    "ScheduleDay",
    "ScheduleParameterDescription",
    "ScheduleParameter",
    "ScheduleNumberDescription",
    "ScheduleNumber",
    "ScheduleSwitchDescription",
    "ScheduleSwitch",
    "SCHEDULE_PARAMETERS",
    "collect_schedule_data",
    "SchedulesStructure",
]



"""Contains an ecoSTER class."""

from __future__ import annotations

from pyplumio.const import DeviceType
from pyplumio.devices import PhysicalDevice, device_handler


@device_handler(DeviceType.ECOSTER)
class EcoSTER(PhysicalDevice):
    """Represents an ecoSTER thermostat."""

    __slots__ = ()


__all__ = ["EcoSTER"]



"""Contains callback filters."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from contextlib import suppress
from copy import copy
from decimal import Decimal
from functools import wraps
import logging
import math
import time
from typing import (
    Any,
    Final,
    Protocol,
    SupportsFloat,
    TypeAlias,
    TypeVar,
    overload,
    runtime_checkable,
)

from pyplumio.helpers.event_manager import EventCallback
from pyplumio.parameters import Number, Parameter

_LOGGER = logging.getLogger(__name__)

numpy_installed = False
with suppress(ImportError):
    import numpy as np

    _LOGGER.info("Using numpy for improved float precision")
    numpy_installed = True


UNDEFINED: Final = "undefined"


@runtime_checkable
class SupportsSubtraction(Protocol):
    """Supports subtraction operation."""

    __slots__ = ()

    def __sub__(
        self: SupportsSubtraction, other: SupportsSubtraction
    ) -> SupportsSubtraction:
        """Subtract a value."""


@runtime_checkable
class SupportsComparison(Protocol):
    """Supports comparison."""

    __slots__ = ()

    def __hash__(self) -> int:
        """Return a hash of the value."""

    def __eq__(self: SupportsComparison, other: SupportsComparison) -> bool:
        """Compare a value."""


_ComparableT = TypeVar("_ComparableT", Parameter, SupportsFloat, SupportsComparison)

DEFAULT_TOLERANCE: Final = 1e-6


@overload
def is_close(old: Parameter, new: Parameter, *, tolerance: None = None) -> bool: ...


@overload
def is_close(
    old: SupportsFloat, new: SupportsFloat, *, tolerance: float = DEFAULT_TOLERANCE
) -> bool: ...


@overload
def is_close(
    old: SupportsComparison, new: SupportsComparison, *, tolerance: None = None
) -> bool: ...


def is_close(
    old: _ComparableT, new: _ComparableT, *, tolerance: float | None = DEFAULT_TOLERANCE
) -> bool:
    """Check if value is significantly changed."""
    if isinstance(new, Parameter) and new.update_pending.is_set():
        return False

    if tolerance and isinstance(old, SupportsFloat) and isinstance(new, SupportsFloat):
        return math.isclose(old, new, abs_tol=tolerance)

    if isinstance(old, Parameter) and isinstance(new, Parameter):
        return old.values.__eq__(new.values)

    return old.__eq__(new)


@overload
def diffence_between(old: list, new: list) -> list: ...


@overload
def diffence_between(
    old: SupportsSubtraction, new: SupportsSubtraction
) -> SupportsSubtraction: ...


def diffence_between(
    old: SupportsSubtraction | list, new: SupportsSubtraction | list
) -> SupportsSubtraction | list | None:
    """Return a difference between values."""
    if isinstance(old, list) and isinstance(new, list):
        return [x for x in new if x not in old]

    if isinstance(old, SupportsSubtraction) and isinstance(new, SupportsSubtraction):
        return new.__sub__(old)

    return None


_Numeric: TypeAlias = float | int | Decimal | Number


def numeric_only(func: Callable) -> Callable:
    """Mark filter as numeric only.

    Ensure that value passed to filter when used as callable is numeric
    otherwise raise TypeError.
    """

    @wraps(func)
    def wrapper(instance: Filter, new_value: _Numeric) -> Any:
        """Wrap __call__ method and add numeric check to it."""
        if not isinstance(new_value, _Numeric):
            raise TypeError(
                f"{instance.__class__.__name__.capitalize()} filter can only be used "
                f"with numeric values, got {type(new_value).__name__}: {new_value}"
            )

        return func(instance, new_value)

    return wrapper


class Filter(ABC):
    """Represents a filter."""

    __slots__ = ("_callback",)

    _callback: EventCallback

    def __init__(self, callback: EventCallback) -> None:
        """Initialize a new filter."""
        self._callback = callback

    def __hash__(self) -> int:
        """Return a hash of the filter based on its callback."""
        return hash(self._callback)

    def __eq__(self, other: Any) -> bool:
        """Compare callbacks."""
        if isinstance(other, Filter):
            return self._callback == other._callback

        if callable(other):
            return bool(self._callback == other)

        return NotImplemented

    @abstractmethod
    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""


class _Aggregate(Filter):
    """Represents an aggregate filter.

    Calls a callback with a sum of values collected over a specified
    time period or when sample size limit reached.
    """

    __slots__ = ("_values", "_sample_size", "_timeout", "_last_call_time")

    _values: list[_Numeric]
    _sample_size: int
    _timeout: float
    _last_call_time: float

    def __init__(
        self, callback: EventCallback, seconds: float, sample_size: int
    ) -> None:
        """Initialize a new aggregate filter."""
        super().__init__(callback)
        self._last_call_time = time.monotonic()
        self._timeout = seconds
        self._sample_size = sample_size
        self._values = []

    @numeric_only
    async def __call__(self, new_value: _Numeric) -> Any:
        """Set a new value for the callback."""
        current_time = time.monotonic()
        self._values.append(new_value)
        time_since_call = current_time - self._last_call_time
        if time_since_call >= self._timeout or len(self._values) >= self._sample_size:
            sum_of_values = (
                np.sum(np.array(self._values)) if numpy_installed else sum(self._values)
            )
            result = await self._callback(float(sum_of_values))
            self._last_call_time = current_time
            self._values = []
            return result


def aggregate(callback: EventCallback, seconds: float, sample_size: int) -> _Aggregate:
    """Create a new aggregate filter.

    A callback function will be called with a sum of values collected
    over a specified time period or when sample size limit reached.
    Can only be used with numeric values.

    :param callback: A callback function to be awaited once filter
        conditions are fulfilled
    :type callback: EventCallback
    :param seconds: A callback will be awaited with a sum of values
        aggregated over this amount of seconds.
    :type seconds: float
    :param sample_size: The maximum number of values to aggregate
        before calling the callback
    :type sample_size: int
    :return: An instance of callable filter
    :rtype: _Aggregate
    """
    return _Aggregate(callback, seconds, sample_size)


class _Clamp(Filter):
    """Represents a clamp filter.

    Calls callback with a value clamped between specified boundaries.
    """

    __slots__ = ("_min_value", "_max_value", "_ignore_out_of_range")

    _min_value: float
    _max_value: float
    _ignore_out_of_range: bool

    def __init__(
        self,
        callback: EventCallback,
        min_value: float,
        max_value: float,
        *,
        ignore_out_of_range: bool = False,
    ) -> None:
        """Initialize a new Clamp filter."""
        super().__init__(callback)
        self._min_value = min_value
        self._max_value = max_value
        self._ignore_out_of_range = ignore_out_of_range

    @numeric_only
    async def __call__(self, new_value: _Numeric) -> Any:
        """Set a new value for the callback."""
        if self._ignore_out_of_range and (
            new_value < self._min_value or new_value > self._max_value
        ):
            return

        if new_value < self._min_value:
            return await self._callback(self._min_value)

        if new_value > self._max_value:
            return await self._callback(self._max_value)

        return await self._callback(new_value)


def clamp(
    callback: EventCallback,
    min_value: float,
    max_value: float,
    *,
    ignore_out_of_range: bool = False,
) -> _Clamp:
    """Create a new clamp filter.

    A callback function will be called and passed value clamped
    between specified boundaries.

    :param callback: A callback function to be awaited on new value
    :type callback: EventCallback
    :param min_value: A lower boundary
    :type min_value: float
    :param max_value: An upper boundary
    :type max_value: float
    :param ignore_out_of_range: If `True`, values outside of
        specified boundaries will be ignored, defaults to `False`
    :type ignore_out_of_range: bool, optional
    :return: An instance of callable filter
    :rtype: _Clamp
    """
    return _Clamp(
        callback, min_value, max_value, ignore_out_of_range=ignore_out_of_range
    )


_FilterFunc: TypeAlias = Callable[[Any], bool]


class _Custom(Filter):
    """Represents a custom filter.

    Calls a callback with value, if user-defined filter function
    that's called by this class with the value as an argument
    returns true.
    """

    __slots__ = ("_filter_func",)

    _filter_func: _FilterFunc

    def __init__(self, callback: EventCallback, filter_func: _FilterFunc) -> None:
        """Initialize a new custom filter."""
        super().__init__(callback)
        self._filter_func = filter_func

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._filter_func(new_value):
            await self._callback(new_value)


def custom(callback: EventCallback, filter_func: _FilterFunc) -> _Custom:
    """Create a new custom filter.

    A callback function will be called when a user-defined filter
    function, that's being called with the value as an argument,
    returns true.

    :param callback: A callback function to be awaited when
        filter function return true
    :type callback: Callback
    :param filter_func: Filter function, that will be called with a
        value and should return `True` to await filter's callback
    :type filter_func: Callable[[Any], bool]
    :return: An instance of callable filter
    :rtype: _Custom
    """
    return _Custom(callback, filter_func)


class _Throttle(Filter):
    """Represents a throttle filter.

    Calls a callback only when certain amount of seconds passed
    since the last call.
    """

    __slots__ = ("_last_called", "_timeout")

    _last_called: float | None
    _timeout: float

    def __init__(self, callback: EventCallback, seconds: float) -> None:
        """Initialize a new throttle filter."""
        super().__init__(callback)
        self._last_called = None
        self._timeout = seconds

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        current_timestamp = time.monotonic()
        if (
            self._last_called is None
            or (current_timestamp - self._last_called) >= self._timeout
        ):
            self._last_called = current_timestamp
            return await self._callback(new_value)


def throttle(callback: EventCallback, seconds: float) -> _Throttle:
    """Create a new throttle filter.

    A callback function will only be called once a certain amount of
    seconds passed since the last call.

    :param callback: A callback function that will be awaited once
        filter conditions are fulfilled
    :type callback: EventCallback
    :param seconds: A callback will be awaited at most once per
        this amount of seconds
    :type seconds: float
    :return: An instance of callable filter
    :rtype: _Throttle
    """
    return _Throttle(callback, seconds)


class ComparisonFilter(Filter):
    """Represents a filter that compares current and previous values."""

    __slots__ = ("_value",)

    _value: Any

    def __init__(self, callback: EventCallback) -> None:
        """Initialize a new comparison filter."""
        self._value = UNDEFINED
        super().__init__(callback)

    def is_undefined(self) -> bool:
        """Check if current value is undefined."""
        return True if self._value == UNDEFINED else False

    @property
    def value(self) -> Any:
        """Return filter value."""
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        """Set filter value."""
        if isinstance(value, Parameter):
            self._value = copy(value)
        else:
            self._value = value


class _Deadband(ComparisonFilter):
    """Represents a deadband filter.

    Calls a callback only when value is significantly changed from the
    previous callback call.
    """

    __slots__ = ("_tolerance",)

    _tolerance: float

    def __init__(self, callback: EventCallback, tolerance: float) -> None:
        """Initialize a new value changed filter."""
        self._tolerance = tolerance
        super().__init__(callback)

    @numeric_only
    async def __call__(self, new_value: _Numeric) -> Any:
        """Set a new value for the callback."""
        if self.is_undefined() or not is_close(
            self.value, new_value, tolerance=self._tolerance
        ):
            self.value = new_value
            return await self._callback(new_value)


def deadband(callback: EventCallback, tolerance: float) -> _Deadband:
    """Create a new deadband filter.

    A callback function will only be called when the value is significantly changed
    from the previous callback call.

    :param callback: A callback function to be awaited on significant value change
    :type callback: EventCallback
    :param tolerance: The minimum difference required to trigger the callback
    :type tolerance: float
    :return: An instance of callable filter
    :rtype: _Deadband
    """
    return _Deadband(callback, tolerance)


class _Debounce(ComparisonFilter):
    """Represents a debounce filter.

    Calls a callback only when value is stabilized across multiple
    filter calls.
    """

    __slots__ = ("_calls", "_min_calls")

    _calls: int
    _min_calls: int

    def __init__(self, callback: EventCallback, min_calls: int) -> None:
        """Initialize a new debounce filter."""
        super().__init__(callback)
        self._calls = 0
        self._min_calls = min_calls

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self.is_undefined() or not is_close(self.value, new_value):
            self._calls += 1
        else:
            self._calls = 0

        if self.is_undefined() or self._calls >= self._min_calls:
            self.value = new_value
            self._calls = 0
            return await self._callback(new_value)


def debounce(callback: EventCallback, min_calls: int) -> _Debounce:
    """Create a new debounce filter.

    A callback function will only be called once the value is stabilized
    across multiple filter calls.

    :param callback: A callback function to be awaited on value change
    :type callback: EventCallback
    :param min_calls: Value shouldn't change for this amount of
        filter calls
    :type min_calls: int
    :return: An instance of callable filter
    :rtype: _Debounce
    """
    return _Debounce(callback, min_calls)


class _Delta(ComparisonFilter):
    """Represents a difference filter.

    Calls a callback with a difference between two subsequent values.
    """

    __slots__ = ()

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self.is_undefined() or not is_close(self.value, new_value):
            old_value = self.value
            self.value = new_value
            if (difference := diffence_between(old_value, new_value)) is not None:
                return await self._callback(difference)


def delta(callback: EventCallback) -> _Delta:
    """Create a new difference filter.

    A callback function will be called with a difference between two
    subsequent values.

    :param callback: A callback function that will be awaited with
        difference between values in two subsequent calls
    :type callback: EventCallback
    :return: An instance of callable filter
    :rtype: _Delta
    """
    return _Delta(callback)


class _OnChange(ComparisonFilter):
    """Represents a value changed filter.

    Calls a callback only when value is changed from the
    previous callback call.
    """

    __slots__ = ()

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self.is_undefined() or not is_close(self.value, new_value):
            self.value = new_value
            return await self._callback(new_value)


def on_change(callback: EventCallback) -> _OnChange:
    """Create a new value changed filter.

    A callback function will only be called if the value is changed from the
    previous call.

    :param callback: A callback function to be awaited on value change
    :type callback: Callback
    :return: An instance of callable filter
    :rtype: _OnChange
    """
    return _OnChange(callback)


__all__ = [
    "Filter",
    "ComparisonFilter",
    "numeric_only",
    "aggregate",
    "clamp",
    "custom",
    "deadband",
    "debounce",
    "delta",
    "on_change",
    "throttle",
]



"""Contains request frames."""

from __future__ import annotations

from typing import Any, cast

from pyplumio.const import (
    ATTR_COUNT,
    ATTR_DEVICE_INDEX,
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SIZE,
    ATTR_START,
    ATTR_SWITCH,
    ATTR_TYPE,
    ATTR_VALUE,
    FrameType,
)
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import Request, Response, frame_handler
from pyplumio.frames.responses import DeviceAvailableResponse, ProgramVersionResponse
from pyplumio.structures.schedules import SCHEDULES
from pyplumio.utils import join_bits


@frame_handler(FrameType.REQUEST_ALERTS)
class AlertsRequest(Request):
    """Represents an alerts request.

    Contains number of alerts to get and index of the first
    alert.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_START, 0), data.get(ATTR_COUNT, 10)])


@frame_handler(FrameType.REQUEST_CHECK_DEVICE)
class CheckDeviceRequest(Request):
    """Represents a check device request."""

    __slots__ = ()

    def response(self, **kwargs: Any) -> Response | None:
        """Return a response frame."""
        return DeviceAvailableResponse(recipient=self.sender, **kwargs)


@frame_handler(FrameType.REQUEST_ECOMAX_CONTROL)
class EcomaxControlRequest(Request):
    """Represents an ecoMAX control request.

    Contains single binary value. 0 - means that controller should
    be turned off, 1 - means that it should be turned on.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray([data[ATTR_VALUE]])
        except KeyError as e:
            raise FrameDataError from e


@frame_handler(FrameType.REQUEST_ECOMAX_PARAMETERS)
class EcomaxParametersRequest(Request):
    """Represents an ecoMAX parameters request.

    Contains number of parameters and index of the first parameter.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


@frame_handler(FrameType.REQUEST_MIXER_PARAMETERS)
class MixerParametersRequest(Request):
    """Represents a mixer parameters request.

    Contains number of parameters to get and index of the first
    parameter.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


@frame_handler(FrameType.REQUEST_PASSWORD)
class PasswordRequest(Request):
    """Represents a password request."""

    __slots__ = ()


@frame_handler(FrameType.REQUEST_PROGRAM_VERSION)
class ProgramVersionRequest(Request):
    """Represents a program version request."""

    __slots__ = ()

    def response(self, **kwargs: Any) -> Response | None:
        """Return a response frame."""
        return ProgramVersionResponse(recipient=self.sender, **kwargs)


@frame_handler(FrameType.REQUEST_REGULATOR_DATA_SCHEMA)
class RegulatorDataSchemaRequest(Request):
    """Represents regulator data schema request."""

    __slots__ = ()


@frame_handler(FrameType.REQUEST_SCHEDULES)
class SchedulesRequest(Request):
    """Represents a schedules request."""

    __slots__ = ()


@frame_handler(FrameType.REQUEST_SET_ECOMAX_PARAMETER)
class SetEcomaxParameterRequest(Request):
    """Represents a request to set an ecoMAX parameter.

    Contains parameter index and new value.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray([data[ATTR_INDEX], data[ATTR_VALUE]])
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


@frame_handler(FrameType.REQUEST_SET_MIXER_PARAMETER)
class SetMixerParameterRequest(Request):
    """Represents a request to set a mixer parameter.

    Contains parameter index, new value and mixer index.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray(
                [data[ATTR_DEVICE_INDEX], data[ATTR_INDEX], data[ATTR_VALUE]]
            )
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


@frame_handler(FrameType.REQUEST_SET_SCHEDULE)
class SetScheduleRequest(Request):
    """Represents a request to set a schedule."""

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        message = b"\1"
        try:
            schedule_type = SCHEDULES.index(data[ATTR_TYPE])
            message += schedule_type.to_bytes(length=1, byteorder="little")
            message += int(data[ATTR_SWITCH]).to_bytes(length=1, byteorder="little")
            message += int(data[ATTR_PARAMETER]).to_bytes(length=1, byteorder="little")
            schedule = cast(list[list[bool]], data[ATTR_SCHEDULE])
        except (KeyError, ValueError) as e:
            raise FrameDataError from e

        return bytearray(message) + bytearray(
            join_bits(day[i : i + 8]) for day in schedule for i in range(0, len(day), 8)
        )


@frame_handler(FrameType.REQUEST_SET_THERMOSTAT_PARAMETER)
class SetThermostatParameterRequest(Request):
    """Represents a request to set a thermostat parameter.

    Contains parameter index, new value, parameter offset and
    parameter size.

    Parameter offset is calculated by multiplying
    thermostat index by total number of parameters available.

    For example, if total available parameters is 12,
    offset for the second thermostat is 12 (1 * 12),
    and for the third thermostat is 24 (2 * 12).
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            index = data[ATTR_INDEX]
            value: int = data[ATTR_VALUE]
            offset = data[ATTR_OFFSET]
            message = bytearray([index if offset is None else index + offset])
            return message + value.to_bytes(length=data[ATTR_SIZE], byteorder="little")
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


@frame_handler(FrameType.REQUEST_START_MASTER)
class StartMasterRequest(Request):
    """Represents a request to become a master.

    Once controller receives this request, it starts sending
    periodic messages.
    """

    __slots__ = ()


@frame_handler(FrameType.REQUEST_STOP_MASTER)
class StopMasterRequest(Request):
    """Represents a request to stop being a master.

    Once controller receives this request, it stops sending
    periodic messages.
    """

    __slots__ = ()


@frame_handler(FrameType.REQUEST_THERMOSTAT_PARAMETERS)
class ThermostatParametersRequest(Request):
    """Represents a thermostat parameters request.

    Contains number of parameters to get and index of the first
    parameter.
    """

    __slots__ = ()

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


@frame_handler(FrameType.REQUEST_UID)
class UIDRequest(Request):
    """Represents an UID request."""

    __slots__ = ()


__all__ = [
    "AlertsRequest",
    "CheckDeviceRequest",
    "EcomaxControlRequest",
    "EcomaxParametersRequest",
    "MixerParametersRequest",
    "PasswordRequest",
    "ProgramVersionRequest",
    "RegulatorDataSchemaRequest",
    "SchedulesRequest",
    "SetEcomaxParameterRequest",
    "SetMixerParameterRequest",
    "SetScheduleRequest",
    "SetThermostatParameterRequest",
    "StartMasterRequest",
    "StopMasterRequest",
    "ThermostatParametersRequest",
    "UIDRequest",
]



"""Contains response frames."""

from __future__ import annotations

from typing import Any

from pyplumio.const import ATTR_PASSWORD, FrameType
from pyplumio.frames import Response, Structured, frame_handler
from pyplumio.structures.alerts import AlertsStructure
from pyplumio.structures.ecomax_parameters import EcomaxParametersStructure
from pyplumio.structures.mixer_parameters import MixerParametersStructure
from pyplumio.structures.network_info import NetworkInfoStructure
from pyplumio.structures.product_info import ProductInfoStructure
from pyplumio.structures.program_version import ProgramVersionStructure
from pyplumio.structures.regulator_data_schema import RegulatorDataSchemaStructure
from pyplumio.structures.schedules import SchedulesStructure
from pyplumio.structures.thermostat_parameters import ThermostatParametersStructure


@frame_handler(FrameType.RESPONSE_ALERTS, structure=AlertsStructure)
class AlertsResponse(Structured, Response):
    """Represents response to a device alerts request."""

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_DEVICE_AVAILABLE, structure=NetworkInfoStructure)
class DeviceAvailableResponse(Structured, Response):
    """Represents a device available response.

    Contains network information and status.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_ECOMAX_CONTROL)
class EcomaxControlResponse(Response):
    """Represents response to an ecoMAX control request.

    Empty response acknowledges, that ecoMAX control request was
    successfully processed.
    """

    __slots__ = ()


@frame_handler(
    FrameType.RESPONSE_ECOMAX_PARAMETERS, structure=EcomaxParametersStructure
)
class EcomaxParametersResponse(Structured, Response):
    """Represents an ecoMAX parameters response.

    Contains editable ecoMAX parameters.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_MIXER_PARAMETERS, structure=MixerParametersStructure)
class MixerParametersResponse(Structured, Response):
    """Represents a mixer parameters response.

    Contains editable mixer parameters.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_PASSWORD)
class PasswordResponse(Response):
    """Represents a password response.

    Contains device service password as plaintext.
    """

    __slots__ = ()

    def decode_message(self, message: bytearray) -> dict[str, Any]:
        """Decode a frame message."""
        password = message[1:].decode() if message[1:] else None
        return {ATTR_PASSWORD: password}


@frame_handler(FrameType.RESPONSE_PROGRAM_VERSION, structure=ProgramVersionStructure)
class ProgramVersionResponse(Structured, Response):
    """Represents a program version response.

    Contains software version info.
    """

    __slots__ = ()


@frame_handler(
    FrameType.RESPONSE_REGULATOR_DATA_SCHEMA, structure=RegulatorDataSchemaStructure
)
class RegulatorDataSchemaResponse(Structured, Response):
    """Represents a regulator data schema response.

    Contains schema, that describes structure of ecoMAX regulator data
    message.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_SCHEDULES, structure=SchedulesStructure)
class SchedulesResponse(Structured, Response):
    """Represents response to a device schedules request."""

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_SET_ECOMAX_PARAMETER)
class SetEcomaxParameterResponse(Response):
    """Represents response to a set ecoMAX parameter request.

    Empty response acknowledges, that ecoMAX parameter was
    successfully set.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_SET_MIXER_PARAMETER)
class SetMixerParameterResponse(Response):
    """Represents response to a set mixer parameter request.

    Empty response acknowledges, that mixer parameter was
    successfully set.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_SET_THERMOSTAT_PARAMETER)
class SetThermostatParameterResponse(Response):
    """Represents response to a set thermostat parameter request.

    Empty response acknowledges, that thermostat parameter was
    successfully set.
    """

    __slots__ = ()


@frame_handler(
    FrameType.RESPONSE_THERMOSTAT_PARAMETERS, structure=ThermostatParametersStructure
)
class ThermostatParametersResponse(Structured, Response):
    """Represents a thermostat parameters response.

    Contains editable thermostat parameters.
    """

    __slots__ = ()


@frame_handler(FrameType.RESPONSE_UID, structure=ProductInfoStructure)
class UIDResponse(Structured, Response):
    """Represents an UID response.

    Contains product info and product UID.
    """

    __slots__ = ()


__all__ = [
    "AlertsResponse",
    "DeviceAvailableResponse",
    "EcomaxControlResponse",
    "EcomaxParametersResponse",
    "MixerParametersResponse",
    "PasswordResponse",
    "ProgramVersionResponse",
    "RegulatorDataSchemaResponse",
    "SchedulesResponse",
    "SetEcomaxParameterResponse",
    "SetMixerParameterResponse",
    "SetThermostatParameterResponse",
    "ThermostatParametersResponse",
    "UIDResponse",
]



"""Contains sensor data decoder."""

from collections.abc import Generator, MutableMapping
from contextlib import suppress
from dataclasses import dataclass
import math
import struct
from typing import Any, Final, TypeVar

from pyplumio.const import ATTR_SCHEDULE, BYTE_UNDEFINED, DeviceState, LambdaState
from pyplumio.data_types import Float, UnsignedInt, UnsignedShort
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_AIR_IN_TEMP: Final = "air_in_temp"
ATTR_AIR_OUT_TEMP: Final = "air_out_temp"
ATTR_ALARM: Final = "alarm"
ATTR_BLOW_FAN1: Final = "blow_fan1"
ATTR_BLOW_FAN2: Final = "blow_fan2"
ATTR_BOILER_LOAD: Final = "boiler_load"
ATTR_BOILER_POWER: Final = "boiler_power"
ATTR_CIRCULATION_PUMP_FLAG: Final = "circulation_pump_flag"
ATTR_CIRCULATION_PUMP: Final = "circulation_pump"
ATTR_CONTACTS: Final = "contacts"
ATTR_CURRENT_TEMP: Final = "current_temp"
ATTR_ECOLAMBDA: Final = "ecolambda"
ATTR_ECOSTER: Final = "ecoster"
ATTR_EXCHANGER_TEMP: Final = "exchanger_temp"
ATTR_EXHAUST_TEMP: Final = "exhaust_temp"
ATTR_FAN_POWER: Final = "fan_power"
ATTR_FAN: Final = "fan"
ATTR_FAN2_EXHAUST: Final = "fan2_exhaust"
ATTR_FEEDER_TEMP: Final = "feeder_temp"
ATTR_FEEDER: Final = "feeder"
ATTR_FEEDER2: Final = "feeder2"
ATTR_FIREPLACE_PUMP: Final = "fireplace_pump"
ATTR_FIREPLACE_TEMP: Final = "fireplace_temp"
ATTR_FUEL_CONSUMPTION: Final = "fuel_consumption"
ATTR_FUEL_LEVEL: Final = "fuel_level"
ATTR_GCZ_CONTACT: Final = "gcz_contact"
ATTR_HEATING_PUMP_FLAG: Final = "heating_pump_flag"
ATTR_HEATING_PUMP: Final = "heating_pump"
ATTR_HEATING_STATUS: Final = "heating_status"
ATTR_HEATING_TARGET: Final = "heating_target"
ATTR_HEATING_TEMP: Final = "heating_temp"
ATTR_HYDRAULIC_COUPLER_TEMP: Final = "hydraulic_coupler_temp"
ATTR_LAMBDA_LEVEL: Final = "lambda_level"
ATTR_LAMBDA_STATE: Final = "lambda_state"
ATTR_LAMBDA_TARGET: Final = "lambda_target"
ATTR_LIGHTER: Final = "lighter"
ATTR_LOWER_BUFFER_TEMP: Final = "lower_buffer_temp"
ATTR_LOWER_SOLAR_TEMP: Final = "lower_solar_temp"
ATTR_MIXER_SENSORS: Final = "mixer_sensors"
ATTR_MIXERS_AVAILABLE: Final = "mixers_available"
ATTR_MIXERS_CONNECTED: Final = "mixers_connected"
ATTR_MODULE_A: Final = "module_a"
ATTR_MODULE_B: Final = "module_b"
ATTR_MODULE_C: Final = "module_c"
ATTR_MODULES: Final = "modules"
ATTR_OPTICAL_TEMP: Final = "optical_temp"
ATTR_OUTER_BOILER: Final = "outer_boiler"
ATTR_OUTER_FEEDER: Final = "outer_feeder"
ATTR_OUTSIDE_TEMP: Final = "outside_temp"
ATTR_PANEL: Final = "panel"
ATTR_PENDING_ALERTS: Final = "pending_alerts"
ATTR_PUMP: Final = "pump"
ATTR_RETURN_TEMP: Final = "return_temp"
ATTR_SOLAR_PUMP_FLAG: Final = "solar_pump_flag"
ATTR_SOLAR_PUMP: Final = "solar_pump"
ATTR_STATE: Final = "state"
ATTR_TARGET_TEMP: Final = "target_temp"
ATTR_THERMOSTAT_SENSORS: Final = "thermostat_sensors"
ATTR_THERMOSTAT: Final = "thermostat"
ATTR_THERMOSTATS_AVAILABLE: Final = "thermostats_available"
ATTR_THERMOSTATS_CONNECTED: Final = "thermostats_connected"
ATTR_TOTAL_GAIN: Final = "total_gain"
ATTR_TRANSMISSION: Final = "transmission"
ATTR_UPPER_BUFFER_TEMP: Final = "upper_buffer_temp"
ATTR_UPPER_SOLAR_TEMP: Final = "upper_solar_temp"
ATTR_WATER_HEATER_PUMP_FLAG: Final = "water_heater_pump_flag"
ATTR_WATER_HEATER_PUMP: Final = "water_heater_pump"
ATTR_WATER_HEATER_STATUS: Final = "water_heater_status"
ATTR_WATER_HEATER_TARGET: Final = "water_heater_target"
ATTR_WATER_HEATER_TEMP: Final = "water_heater_temp"

OUTPUTS: tuple[str, ...] = (
    ATTR_FAN,
    ATTR_FEEDER,
    ATTR_HEATING_PUMP,
    ATTR_WATER_HEATER_PUMP,
    ATTR_CIRCULATION_PUMP,
    ATTR_LIGHTER,
    ATTR_ALARM,
    ATTR_OUTER_BOILER,
    ATTR_FAN2_EXHAUST,
    ATTR_FEEDER2,
    ATTR_OUTER_FEEDER,
    ATTR_SOLAR_PUMP,
    ATTR_FIREPLACE_PUMP,
    ATTR_GCZ_CONTACT,
    ATTR_BLOW_FAN1,
    ATTR_BLOW_FAN2,
)

TEMPERATURES: tuple[str, ...] = (
    ATTR_HEATING_TEMP,
    ATTR_FEEDER_TEMP,
    ATTR_WATER_HEATER_TEMP,
    ATTR_OUTSIDE_TEMP,
    ATTR_RETURN_TEMP,
    ATTR_EXHAUST_TEMP,
    ATTR_OPTICAL_TEMP,
    ATTR_UPPER_BUFFER_TEMP,
    ATTR_LOWER_BUFFER_TEMP,
    ATTR_UPPER_SOLAR_TEMP,
    ATTR_LOWER_SOLAR_TEMP,
    ATTR_FIREPLACE_TEMP,
    ATTR_TOTAL_GAIN,
    ATTR_HYDRAULIC_COUPLER_TEMP,
    ATTR_EXCHANGER_TEMP,
    ATTR_AIR_IN_TEMP,
    ATTR_AIR_OUT_TEMP,
)

STATUSES: tuple[str, ...] = (
    ATTR_HEATING_TARGET,
    ATTR_HEATING_STATUS,
    ATTR_WATER_HEATER_TARGET,
    ATTR_WATER_HEATER_STATUS,
)

MODULES: tuple[str, ...] = (
    ATTR_MODULE_A,
    ATTR_MODULE_B,
    ATTR_MODULE_C,
    ATTR_ECOLAMBDA,
    ATTR_ECOSTER,
    ATTR_PANEL,
)

FUEL_LEVEL_OFFSET: Final = 101


@dataclass(slots=True, kw_only=True)
class ConnectedModules:
    """Represents a firmware version info for connected module."""

    module_a: str | None = None
    module_b: str | None = None
    module_c: str | None = None
    ecolambda: str | None = None
    ecoster: str | None = None
    panel: str | None = None


struct_version = struct.Struct("<BBB")
struct_vendor = struct.Struct("<BB")

_DataT = TypeVar("_DataT", bound=MutableMapping)


class SensorDataStructure(StructureDecoder):
    """Represents a sensor data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _decode_outputs(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode outputs from message."""
        outputs = UnsignedInt.from_bytes(message, self._offset)
        self._offset += outputs.size
        for index, output in enumerate(OUTPUTS):
            data[output] = bool(outputs.value & 2**index)

        return data

    def _decode_output_flags(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode output flags from message."""
        output_flags = UnsignedInt.from_bytes(message, self._offset)
        self._offset += output_flags.size
        data[ATTR_HEATING_PUMP_FLAG] = bool(output_flags.value & 0x04)
        data[ATTR_WATER_HEATER_PUMP_FLAG] = bool(output_flags.value & 0x08)
        data[ATTR_CIRCULATION_PUMP_FLAG] = bool(output_flags.value & 0x10)
        data[ATTR_SOLAR_PUMP_FLAG] = bool(output_flags.value & 0x800)
        return data

    def _decode_temperatures(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode temperatures from message."""
        offset = self._offset
        temperatures = message[offset]
        offset += 1
        for _ in range(temperatures):
            index = message[offset]
            offset += 1
            temp = Float.from_bytes(message, offset)
            offset += temp.size
            if (not math.isnan(temp.value)) and 0 <= index < len(TEMPERATURES):
                # Temperature exists and index is in the correct range.
                data[TEMPERATURES[index]] = temp.value

        self._offset = offset
        return data

    def _decode_statuses(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode statuses from message."""
        for index, status in enumerate(STATUSES):
            data[status] = message[self._offset + index]

        self._offset += len(STATUSES)
        return data

    def _decode_pending_alerts(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode pending alerts from message."""
        pending_alerts = message[self._offset]
        data[ATTR_PENDING_ALERTS] = pending_alerts
        self._offset += pending_alerts + 1
        return data

    def _decode_fuel_level(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode fuel level from message."""
        fuel_level = message[self._offset]
        self._offset += 1
        if fuel_level != BYTE_UNDEFINED:
            # Fuel offset requirement on at least ecoMAX 860P6-O.
            # See: https://github.com/denpamusic/PyPlumIO/issues/19
            data[ATTR_FUEL_LEVEL] = (
                fuel_level
                if fuel_level < FUEL_LEVEL_OFFSET
                else fuel_level - FUEL_LEVEL_OFFSET
            )

        return data

    def _decode_boiler_load(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode boiler load from message."""
        boiler_load = message[self._offset]
        self._offset += 1
        if boiler_load != BYTE_UNDEFINED:
            data[ATTR_BOILER_LOAD] = boiler_load

        return data

    def _decode_float_value(
        self, name: str, message: bytearray, data: _DataT
    ) -> _DataT:
        """Decode float value and increase an offset."""
        float_value = Float.from_bytes(message, self._offset)
        self._offset += float_value.size
        if not math.isnan(float_value.value):
            data[name] = float_value.value

        return data

    def _decode_modules(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode modules from message."""
        offset = self._offset

        def _module_versions() -> Generator[tuple[str, str | None]]:
            """Unpack a module version."""
            nonlocal offset
            for module in MODULES:
                if message[offset] != BYTE_UNDEFINED:
                    version_data = struct_version.unpack_from(message, offset)
                    version = ".".join(str(i) for i in version_data)
                    offset += struct_version.size
                    if module == ATTR_MODULE_A:
                        vendor_code, vendor_version = struct_vendor.unpack_from(
                            message, offset
                        )
                        version += f".{chr(vendor_code) + str(vendor_version)}"
                        offset += struct_vendor.size
                else:
                    offset += 1
                    version = None

                yield module, version

        data[ATTR_MODULES] = ConnectedModules(**dict(_module_versions()))
        self._offset = offset
        return data

    def _decode_lambda_sensor(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode lambda sensor from message."""
        offset = self._offset
        lambda_state = message[offset]
        offset += 1
        if lambda_state != BYTE_UNDEFINED:
            lambda_target = message[offset]
            offset += 1
            level = UnsignedShort.from_bytes(message, offset)
            offset += level.size
            with suppress(ValueError):
                lambda_state = LambdaState(lambda_state)

            data[ATTR_LAMBDA_STATE] = lambda_state
            data[ATTR_LAMBDA_TARGET] = lambda_target
            data[ATTR_LAMBDA_LEVEL] = level.value / 10

        self._offset = offset
        return data

    def _decode_thermostat_sensors(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode thermostat sensors from message."""
        contact_mask = 1
        schedule_mask = 1 << 3
        offset = self._offset

        def _unpack_thermostat_sensors(contacts: int) -> dict[str, Any] | None:
            """Unpack sensors for a single thermostat."""
            nonlocal offset, contact_mask, schedule_mask
            state = message[offset]
            offset += 1
            current_temp = Float.from_bytes(message, offset)
            offset += current_temp.size
            target_temp = Float.from_bytes(message, offset)
            offset += target_temp.size
            contacts_state = bool(contacts & contact_mask)
            contact_mask <<= 1
            schedule_state = bool(contacts & schedule_mask)
            schedule_mask <<= 1

            if math.isnan(current_temp.value) or target_temp.value <= 0:
                return None

            return {
                ATTR_STATE: state,
                ATTR_CURRENT_TEMP: current_temp.value,
                ATTR_TARGET_TEMP: target_temp.value,
                ATTR_CONTACTS: contacts_state,
                ATTR_SCHEDULE: schedule_state,
            }

        def _thermostat_sensors(contacts: int) -> Generator[tuple[int, dict[str, Any]]]:
            """Get thermostat sensors."""
            for index in range(thermostats):
                if sensors := _unpack_thermostat_sensors(contacts):
                    yield (index, sensors)

        contacts = message[offset]
        offset += 1
        if contacts != BYTE_UNDEFINED:
            thermostats = message[offset]
            offset += 1
            thermostat_sensors = dict(_thermostat_sensors(contacts))
            data[ATTR_THERMOSTAT_SENSORS] = thermostat_sensors
            data[ATTR_THERMOSTATS_CONNECTED] = len(thermostat_sensors)
            data[ATTR_THERMOSTATS_AVAILABLE] = thermostats

        self._offset = offset
        return data

    def _decode_mixer_sensors(self, message: bytearray, data: _DataT) -> _DataT:
        """Decode mixer sensors from message."""
        offset = self._offset

        def _unpack_mixer_sensors() -> dict[str, Any] | None:
            """Unpack sensors for a single mixer."""
            nonlocal offset
            current_temp = Float.from_bytes(message, offset)
            offset += current_temp.size
            data = None
            if not math.isnan(current_temp.value):
                data = {
                    ATTR_CURRENT_TEMP: current_temp.value,
                    ATTR_TARGET_TEMP: message[offset],
                    ATTR_PUMP: bool(message[offset + 2] & 0x01),
                }

            offset += 4
            return data

        def _mixer_sensors(mixers: int) -> Generator[tuple[int, dict[str, Any]]]:
            """Get mixer sensors."""
            for index in range(mixers):
                if sensors := _unpack_mixer_sensors():
                    yield (index, sensors)

        mixers = message[offset]
        offset += 1
        mixer_sensors = dict(_mixer_sensors(mixers))
        data[ATTR_MIXER_SENSORS] = mixer_sensors
        data[ATTR_MIXERS_CONNECTED] = len(mixer_sensors)
        data[ATTR_MIXERS_AVAILABLE] = mixers
        self._offset = offset
        return data

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        data = ensure_dict(data)
        data[ATTR_STATE] = message[offset]
        self._offset = offset + 1
        with suppress(ValueError):
            data[ATTR_STATE] = DeviceState(data[ATTR_STATE])

        data = self._decode_outputs(message, data)
        data = self._decode_output_flags(message, data)
        data = self._decode_temperatures(message, data)
        data = self._decode_statuses(message, data)
        data = self._decode_pending_alerts(message, data)
        data = self._decode_fuel_level(message, data)
        data[ATTR_TRANSMISSION] = message[self._offset]
        self._offset += 1
        data = self._decode_float_value(ATTR_FAN_POWER, message, data)
        data = self._decode_boiler_load(message, data)
        data = self._decode_float_value(ATTR_BOILER_POWER, message, data)
        data = self._decode_float_value(ATTR_FUEL_CONSUMPTION, message, data)
        data[ATTR_THERMOSTAT] = message[self._offset]
        self._offset += 1
        data = self._decode_modules(message, data)
        data = self._decode_lambda_sensor(message, data)
        data = self._decode_thermostat_sensors(message, data)
        data = self._decode_mixer_sensors(message, data)
        return data, offset


__all__ = [
    "ATTR_AIR_IN_TEMP",
    "ATTR_AIR_OUT_TEMP",
    "ATTR_ALARM",
    "ATTR_BLOW_FAN1",
    "ATTR_BLOW_FAN2",
    "ATTR_BOILER_LOAD",
    "ATTR_BOILER_POWER",
    "ATTR_CIRCULATION_PUMP",
    "ATTR_CIRCULATION_PUMP_FLAG",
    "ATTR_CONTACTS",
    "ATTR_CURRENT_TEMP",
    "ATTR_ECOLAMBDA",
    "ATTR_ECOSTER",
    "ATTR_EXCHANGER_TEMP",
    "ATTR_EXHAUST_TEMP",
    "ATTR_FAN",
    "ATTR_FAN2_EXHAUST",
    "ATTR_FAN_POWER",
    "ATTR_FEEDER",
    "ATTR_FEEDER2",
    "ATTR_FEEDER_TEMP",
    "ATTR_FIREPLACE_PUMP",
    "ATTR_FIREPLACE_TEMP",
    "ATTR_FUEL_CONSUMPTION",
    "ATTR_FUEL_LEVEL",
    "ATTR_GCZ_CONTACT",
    "ATTR_HEATING_PUMP",
    "ATTR_HEATING_PUMP_FLAG",
    "ATTR_HEATING_STATUS",
    "ATTR_HEATING_TARGET",
    "ATTR_HEATING_TEMP",
    "ATTR_HYDRAULIC_COUPLER_TEMP",
    "ATTR_LAMBDA_LEVEL",
    "ATTR_LAMBDA_STATE",
    "ATTR_LAMBDA_TARGET",
    "ATTR_LIGHTER",
    "ATTR_LOWER_BUFFER_TEMP",
    "ATTR_LOWER_SOLAR_TEMP",
    "ATTR_MIXER_SENSORS",
    "ATTR_MIXERS_AVAILABLE",
    "ATTR_MIXERS_CONNECTED",
    "ATTR_MODULE_A",
    "ATTR_MODULE_B",
    "ATTR_MODULE_C",
    "ATTR_MODULES",
    "ATTR_OPTICAL_TEMP",
    "ATTR_OUTER_BOILER",
    "ATTR_OUTER_FEEDER",
    "ATTR_OUTSIDE_TEMP",
    "ATTR_PANEL",
    "ATTR_PENDING_ALERTS",
    "ATTR_PUMP",
    "ATTR_RETURN_TEMP",
    "ATTR_SOLAR_PUMP",
    "ATTR_SOLAR_PUMP_FLAG",
    "ATTR_STATE",
    "ATTR_TARGET_TEMP",
    "ATTR_THERMOSTAT",
    "ATTR_THERMOSTAT_SENSORS",
    "ATTR_THERMOSTATS_AVAILABLE",
    "ATTR_THERMOSTATS_CONNECTED",
    "ATTR_TOTAL_GAIN",
    "ATTR_TRANSMISSION",
    "ATTR_UPPER_BUFFER_TEMP",
    "ATTR_UPPER_SOLAR_TEMP",
    "ATTR_WATER_HEATER_PUMP",
    "ATTR_WATER_HEATER_PUMP_FLAG",
    "ATTR_WATER_HEATER_STATUS",
    "ATTR_WATER_HEATER_TARGET",
    "ATTR_WATER_HEATER_TEMP",
    "ConnectedModules",
    "MODULES",
    "OUTPUTS",
    "SensorDataStructure",
    "STATUSES",
    "TEMPERATURES",
]



"""Contains message frames."""

from __future__ import annotations

from pyplumio.const import ATTR_SENSORS, FrameType
from pyplumio.frames import Message, Structured, contains, frame_handler
from pyplumio.structures.frame_versions import FrameVersionsStructure
from pyplumio.structures.regulator_data import RegulatorDataStructure
from pyplumio.structures.sensor_data import SensorDataStructure


@frame_handler(FrameType.MESSAGE_REGULATOR_DATA, structure=RegulatorDataStructure)
class RegulatorDataMessage(Structured, Message):
    """Represents a regulator data message."""

    __slots__ = ()


@frame_handler(FrameType.MESSAGE_SENSOR_DATA)
@contains(FrameVersionsStructure, SensorDataStructure, container=ATTR_SENSORS)
class SensorDataMessage(Structured, Message):
    """Represents a sensor data message."""

    __slots__ = ()


__all__ = ["RegulatorDataMessage", "SensorDataMessage"]



"""Contains tests for the protocol classes."""

import asyncio
from dataclasses import asdict
from datetime import datetime, timedelta
import logging
from typing import cast
from unittest.mock import AsyncMock, Mock, PropertyMock, call, patch

import pytest

from pyplumio.const import ATTR_CONNECTED, DeviceType, EncryptionType
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.exceptions import (
    FrameDataError,
    ProtocolError,
    ReadError,
    UnknownDeviceError,
    UnknownFrameError,
)
from pyplumio.frames import Request, Response
from pyplumio.frames.requests import (
    CheckDeviceRequest,
    ProgramVersionRequest,
    StartMasterRequest,
)
from pyplumio.protocol import AsyncProtocol, DummyProtocol, Queues, Statistics
from pyplumio.stream import FrameReader, FrameWriter
from pyplumio.structures.network_info import (
    EthernetParameters,
    NetworkInfo,
    WirelessParameters,
)
from pyplumio.structures.regulator_data import ATTR_REGDATA


@pytest.fixture(name="skip_asyncio_create_task")
def fixture_skip_asyncio_create_task():
    """Bypass asyncio create task."""
    with patch("asyncio.create_task"):
        yield


@pytest.fixture(name="async_protocol")
def fixture_async_protocol() -> AsyncProtocol:
    """Return an async protocol instance."""
    return AsyncProtocol(
        wireless_parameters=WirelessParameters(
            ssid="test_ssid",
            ip="2.2.2.2",
            encryption=EncryptionType.WPA2,
            netmask="255.255.255.255",
            gateway="2.2.2.1",
            signal_quality=50,
            status=True,
        ),
        ethernet_parameters=EthernetParameters(
            ip="1.1.1.2",
            netmask="255.255.255.255",
            gateway="1.1.1.1",
            status=True,
        ),
    )


async def test_dummy_protocol() -> None:
    """Test a dummy protocol."""
    mock_on_connection_lost = AsyncMock()
    dummy_protocol = DummyProtocol()
    dummy_protocol.on_connection_lost.add(mock_on_connection_lost)

    mock_reader = AsyncMock(spec=asyncio.StreamReader)
    mock_writer = AsyncMock(spec=asyncio.StreamWriter, new_callable=Mock)

    # Test establishing the connection.
    with (
        patch("pyplumio.protocol.FrameReader") as mock_frame_reader,
        patch("pyplumio.protocol.FrameWriter") as mock_frame_writer,
    ):
        dummy_protocol.connection_established(mock_reader, mock_writer)
        mock_frame_reader.assert_called_once_with(mock_reader)
        mock_frame_writer.assert_called_once_with(mock_writer)

    assert dummy_protocol.connected.is_set()

    # Test losing the connection.
    with patch(
        "pyplumio.protocol.Protocol.close_writer", new_callable=AsyncMock
    ) as mock_close_writer:
        await dummy_protocol.connection_lost()
        mock_close_writer.assert_awaited_once()

    assert not dummy_protocol.connected.is_set()
    mock_on_connection_lost.assert_awaited_once()

    # Test shutting down the connection.
    with (
        patch("asyncio.Event.is_set", return_value=True),
        patch(
            "pyplumio.protocol.Protocol.close_writer", new_callable=AsyncMock
        ) as mock_close_writer,
    ):
        await dummy_protocol.shutdown()
        mock_close_writer.assert_awaited_once()

    assert not dummy_protocol.connected.is_set()


@pytest.mark.usefixtures("frozen_time")
def test_statistics() -> None:
    """Test statistics dataclass."""
    statistics = Statistics(failed_frames=1)
    assert statistics.connected_since == "never"
    assert statistics.failed_frames == 1
    assert statistics.connection_losses == 0
    assert statistics.connection_loss_at == "never"
    assert statistics.sent_bytes == 0
    assert statistics.sent_frames == 0
    assert statistics.received_bytes == 0
    assert statistics.received_frames == 0
    statistics.update_sent(Request())
    assert statistics.sent_bytes == 10
    assert statistics.sent_frames == 1
    statistics.update_received(Response())
    assert statistics.received_bytes == 10
    assert statistics.received_frames == 1
    statistics.reset_transfer_statistics()
    assert statistics.sent_bytes == 0
    assert statistics.sent_frames == 0
    assert statistics.received_bytes == 0
    assert statistics.received_frames == 0
    assert statistics.failed_frames == 0
    statistics.update_connection_lost()
    assert statistics.connection_losses == 1
    assert statistics.connection_loss_at == datetime.now()


@patch("pyplumio.protocol.AsyncProtocol.create_task")
@patch("pyplumio.protocol.AsyncProtocol.frame_consumer", new_callable=Mock)
@patch("pyplumio.protocol.AsyncProtocol.frame_producer", new_callable=Mock)
@pytest.mark.usefixtures("frozen_time")
async def test_async_protocol_connection_established(
    mock_frame_producer, mock_frame_consumer, mock_create_task
) -> None:
    """Test establishing connection with an async protocol."""
    with patch("asyncio.Queue"):
        async_protocol = AsyncProtocol()

    # Create stream reader, stream writer and queue mocks..
    mock_stream_reader = Mock(spec=asyncio.StreamReader)
    mock_stream_writer = Mock(spec=asyncio.StreamWriter)
    mock_read_queue = Mock(spec=asyncio.Queue)
    mock_write_queue = Mock(spec=asyncio.Queue)
    mock_put_nowait = mock_write_queue.put_nowait

    # Create ecoMAX device mock and add it to the protocol.
    mock_ecomax = Mock(spec=EcoMAX, new_callable=Mock)
    await async_protocol.dispatch("ecomax", mock_ecomax)

    # Test connection established.
    with (
        patch.object(
            async_protocol, "_queues", Queues(mock_read_queue, mock_write_queue)
        ),
        patch(
            "pyplumio.protocol.Statistics.reset_transfer_statistics"
        ) as mock_reset_transfer_statistics,
    ):
        async_protocol.connection_established(mock_stream_reader, mock_stream_writer)

    # Check that StartMaster request was added to the write queue.
    mock_put_nowait.assert_called_once()
    frame = mock_put_nowait.call_args.args[0]
    assert isinstance(frame, StartMasterRequest)
    assert frame.recipient == 69

    # Check that two frame consumers, one write consumer and one frame
    # producer were created.
    assert mock_frame_consumer.call_count == 3
    mock_frame_producer.assert_called_once()
    assert mock_create_task.call_count == 4

    # Check that devices were notified.
    mock_ecomax.dispatch_nowait.assert_called_once_with(ATTR_CONNECTED, True)

    # Check statistics.
    mock_reset_transfer_statistics.assert_called_once()
    assert async_protocol.statistics.connected_since == datetime.now()


async def test_async_protocol_connection_lost() -> None:
    """Test losing the connection with an async protocol."""
    mock_read_queue = Mock(spec=asyncio.Queue)
    mock_write_queue = Mock(spec=asyncio.Queue)

    # Create connection lost callback mock.
    mock_on_connection_lost = AsyncMock()
    async_protocol = AsyncProtocol()
    async_protocol.on_connection_lost.add(mock_on_connection_lost)

    # Create ecoMAX device mock and add it to the protocol.
    mock_ecomax = Mock(spec=EcoMAX, new_callable=AsyncMock)
    mock_writer = AsyncMock(spec=FrameWriter)
    async_protocol.writer = mock_writer
    await async_protocol.dispatch("ecomax", mock_ecomax)

    # Ensure that on connection lost callback not awaited, when
    # connection is not established.
    await async_protocol.connection_lost()
    async_protocol.connected.set()
    mock_on_connection_lost.assert_not_awaited()

    with patch.object(
        async_protocol, "_queues", Queues(mock_read_queue, mock_write_queue)
    ):
        await async_protocol.connection_lost()

    # Check that devices were notified and callback was called.
    mock_ecomax.dispatch.assert_called_once_with(ATTR_CONNECTED, False)
    mock_on_connection_lost.assert_called_once()

    # Check that writer was closed.
    mock_writer.close.assert_awaited_once()
    assert async_protocol.writer is None


@patch("asyncio.wait_for")
@patch("asyncio.gather", new_callable=AsyncMock)
@patch("pyplumio.protocol.AsyncProtocol.cancel_tasks")
@patch("pyplumio.devices.ecomax.EcoMAX.shutdown", new_callable=Mock)
@patch("pyplumio.helpers.event_manager.EventManager.dispatch", new_callable=Mock)
@patch.object(
    AsyncProtocol,
    "data",
    {"ecomax": EcoMAX(queue=asyncio.Queue(), network=NetworkInfo())},
)
@pytest.mark.usefixtures("skip_asyncio_events")
async def test_async_protocol_shutdown(
    mock_dispatch,
    mock_shutdown,
    mock_cancel_tasks,
    mock_gather,
    mock_wait_for,
    async_protocol: AsyncProtocol,
) -> None:
    """Test shutting down connection with an async protocol."""
    mock_read_queue = Mock()
    mock_write_queue = Mock()

    mock_writer = AsyncMock(spec=FrameWriter)
    mock_writer.close = AsyncMock()
    async_protocol.writer = mock_writer

    mock_frame_consumer_task = Mock(spec=asyncio.Task)
    mock_frame_producer_task = Mock(spec=asyncio.Task)

    with (
        patch.object(
            async_protocol, "_queues", Queues(mock_read_queue, mock_write_queue)
        ),
        patch(
            "pyplumio.protocol.AsyncProtocol.tasks",
            return_value=[
                mock_frame_consumer_task,
                mock_frame_producer_task,
            ],
            new_callable=PropertyMock,
        ),
    ):
        await async_protocol.shutdown()

    mock_cancel_tasks.assert_called_once()
    mock_gather.await_count = 3
    mock_gather.assert_has_awaits(
        [
            call(mock_read_queue.join(), mock_write_queue.join()),
            call(*async_protocol.tasks, return_exceptions=True),
            call(mock_dispatch()),
            call(mock_shutdown()),
        ]
    )
    mock_writer.close.assert_awaited_once()
    assert async_protocol.writer is None


@pytest.mark.usefixtures(
    "skip_asyncio_events", "skip_asyncio_create_task", "frozen_time"
)
async def test_async_protocol_frame_producer(
    async_protocol: AsyncProtocol, caplog
) -> None:
    """Test a frame producer task within an async protocol."""
    success = Response(sender=DeviceType.ECOMAX)
    responses = (
        success,
        ProtocolError("test protocol error"),
        UnknownDeviceError("test unknown device error"),
        UnknownFrameError("test unknown frame error"),
        ReadError("test read error"),
        FrameDataError("test frame data error"),
        Exception("test generic error"),
        ConnectionError,
    )

    # Create mock frame reader and writer.
    mock_reader = AsyncMock(spec=FrameReader)
    mock_reader.read = AsyncMock(side_effect=responses)
    mock_writer = AsyncMock(spec=FrameWriter)

    # Create mock queues.
    mock_read_queue = AsyncMock(spec=asyncio.Queue)
    mock_write_queue = AsyncMock(spec=asyncio.Queue)
    mock_write_queue.empty = Mock(
        side_effect=(
            False,
            *(True for _ in range(len(responses) - 1)),
        )
    )
    request = Request()
    mock_write_queue.get = AsyncMock(return_value=request)

    with (
        patch("pyplumio.devices.ecomax.EcoMAX"),
        patch(
            "pyplumio.protocol.AsyncProtocol.connection_lost", new_callable=Mock
        ) as mock_connection_lost,
        caplog.at_level(logging.DEBUG),
    ):
        await async_protocol.frame_producer(
            Queues(mock_read_queue, mock_write_queue),
            reader=mock_reader,
            writer=mock_writer,
        )

    assert caplog.record_tuples == [
        (
            "pyplumio.protocol",
            logging.DEBUG,
            "Can't process received frame: test protocol error",
        ),
        (
            "pyplumio.protocol",
            logging.DEBUG,
            "Can't process received frame: test unknown device error",
        ),
        (
            "pyplumio.protocol",
            logging.DEBUG,
            "Can't process received frame: test unknown frame error",
        ),
        (
            "pyplumio.protocol",
            logging.DEBUG,
            "Can't process received frame: test read error",
        ),
        (
            "pyplumio.protocol",
            logging.DEBUG,
            "Can't process received frame: test frame data error",
        ),
        (
            "pyplumio.protocol",
            logging.ERROR,
            "Unexpected exception",
        ),
    ]

    mock_writer.write.assert_awaited_once_with(request)
    mock_write_queue.task_done.assert_called_once()
    mock_read_queue.put_nowait.assert_called_once_with(success)
    mock_connection_lost.assert_called_once()
    assert mock_write_queue.get.await_count == 1
    assert mock_write_queue.empty.call_count == 8
    assert mock_reader.read.await_count == 8

    # Check statistics.
    assert asdict(async_protocol.statistics) == {
        "connected_since": "never",
        "received_bytes": 10,
        "received_frames": 1,
        "sent_bytes": 10,
        "sent_frames": 1,
        "failed_frames": 5,
        "connection_losses": 1,
        "connection_loss_at": datetime.now(),
        "devices": set(),
    }


@patch("pyplumio.frames.requests.CheckDeviceRequest.response")
@patch("pyplumio.frames.requests.ProgramVersionRequest.response")
@pytest.mark.usefixtures("skip_asyncio_events")
async def test_async_protocol_frame_consumer(
    mock_program_version_response,
    mock_device_available_response,
    async_protocol: AsyncProtocol,
    caplog,
    frozen_time,
) -> None:
    """Test a frame consumer task within an async protocol."""
    mock_read_queue = Mock(spec=asyncio.Queue)
    mock_write_queue = Mock(spec=asyncio.Queue)
    mock_read_queue.get = AsyncMock()

    with (
        patch.object(
            async_protocol, "_queues", Queues(mock_read_queue, mock_write_queue)
        ),
        patch("asyncio.Lock.acquire", return_value=True) as mock_lock_acquire,
        patch("asyncio.Lock.release") as mock_lock_release,
    ):
        await async_protocol.get_device_entry(DeviceType.ECOMAX)

    mock_lock_acquire.assert_awaited_once()
    mock_lock_release.assert_called_once()

    mock_read_queue.get.side_effect = (
        CheckDeviceRequest(sender=DeviceType.ECOMAX),
        ProgramVersionRequest(sender=DeviceType.ECOMAX),
    )

    with (
        caplog.at_level(logging.DEBUG),
        patch("asyncio.Event.is_set", side_effect=(True, True, False)),
    ):
        await async_protocol.frame_consumer(mock_read_queue)

    await async_protocol.shutdown()

    # Check that network settings is correctly set.
    mock_device_available_response.assert_called_once_with(
        data={
            "network": NetworkInfo(
                eth=EthernetParameters(
                    ip="1.1.1.2",
                    netmask="255.255.255.255",
                    gateway="1.1.1.1",
                    status=True,
                ),
                wlan=WirelessParameters(
                    ip="2.2.2.2",
                    netmask="255.255.255.255",
                    gateway="2.2.2.1",
                    status=True,
                    ssid="test_ssid",
                    encryption=EncryptionType.WPA2,
                    signal_quality=50,
                ),
                server_status=True,
            )
        }
    )

    mock_write_queue.put_nowait.assert_has_calls(
        [
            call(mock_device_available_response()),
            call(mock_program_version_response()),
        ]
    )
    assert mock_read_queue.task_done.call_count == 2

    # Test statistics.
    device_statistics = async_protocol.statistics.devices.pop()
    ecomax = cast(EcoMAX, async_protocol.get_nowait("ecomax"))
    await ecomax.dispatch(ATTR_REGDATA, True)
    assert device_statistics.last_seen == datetime.now()
    frozen_time.tick(timedelta(seconds=10))
    await ecomax.dispatch(ATTR_REGDATA, True)
    assert device_statistics.last_seen == datetime.now()
